"use strict";

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.default = void 0;

function _helperOptimiseCallExpression() {
  const data = _interopRequireDefault(require("@babel/helper-optimise-call-expression"));

  _helperOptimiseCallExpression = function () {
    return data;
  };

  return data;
}

function t() {
  const data = _interopRequireWildcard(require("@babel/types"));

  t = function () {
    return data;
  };

  return data;
}

function _interopRequireWildcard(obj) { if (obj && obj.__esModule) { return obj; } else { var newObj = {}; if (obj != null) { for (var key in obj) { if (Object.prototype.hasOwnProperty.call(obj, key)) { var desc = Object.defineProperty && Object.getOwnPropertyDescriptor ? Object.getOwnPropertyDescriptor(obj, key) : {}; if (desc.get || desc.set) { Object.defineProperty(newObj, key, desc); } else { newObj[key] = obj[key]; } } } } newObj.default = obj; return newObj; } }

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

const HARDCORE_THIS_REF = new WeakSet();

function isIllegalBareSuper(node, parent) {
  if (!t().isSuper(node)) return false;
  if (t().isMemberExpression(parent, {
    computed: false
  })) return false;
  if (t().isCallExpression(parent, {
    callee: node
  })) return false;
  return true;
}

function isMemberExpressionSuper(node) {
  return t().isMemberExpression(node) && t().isSuper(node.object);
}

function getPrototypeOfExpression(objectRef, isStatic) {
  const targetRef = isStatic ? objectRef : t().memberExpression(objectRef, t().identifier("prototype"));
  return t().logicalExpression("||", t().memberExpression(targetRef, t().identifier("__proto__")), t().callExpression(t().memberExpression(t().identifier("Object"), t().identifier("getPrototypeOf")), [t().cloneNode(targetRef)]));
}

const visitor = {
  Function(path) {
    if (!path.isArrowFunctionExpression()) path.skip();
  },

  ClassProperty(path) {
    if (!path.node.static) path.skip();
  },

  ReturnStatement(path, state) {
    if (!path.getFunctionParent().isArrowFunctionExpression()) {
      state.returns.push(path);
    }
  },

  ThisExpression(path, state) {
    if (!HARDCORE_THIS_REF.has(path.node)) {
      state.thises.push(path);
    }
  },

  enter(path, state) {
    let callback = state.specHandle;
    if (state.isLoose) callback = state.looseHandle;
    const isBareSuper = path.isCallExpression() && path.get("callee").isSuper();
    const result = callback.call(state, path);

    if (result) {
      state.hasSuper = true;
    }

    if (isBareSuper) {
      state.bareSupers.push(path);
    }

    if (result === true) {
      path.requeue();
    }

    if (result !== true && result) {
      if (Array.isArray(result)) {
        path.replaceWithMultiple(result);
      } else {
        path.replaceWith(result);
      }
    }
  }

};

class ReplaceSupers {
  constructor(opts, inClass = false) {
    this.forceSuperMemoisation = opts.forceSuperMemoisation;
    this.methodPath = opts.methodPath;
    this.methodNode = opts.methodNode;
    this.superRef = opts.superRef;
    this.isStatic = opts.isStatic;
    this.hasSuper = false;
    this.inClass = inClass;
    this.inConstructor = opts.inConstructor;
    this.isLoose = opts.isLoose;
    this.scope = this.methodPath.scope;
    this.file = opts.file;
    this.opts = opts;
    this.bareSupers = [];
    this.returns = [];
    this.thises = [];
  }

  getObjectRef() {
    return t().cloneNode(this.opts.objectRef || this.opts.getObjectRef());
  }

  setSuperProperty(property, value, isComputed) {
    return t().callExpression(this.file.addHelper("set"), [getPrototypeOfExpression(this.getObjectRef(), this.isStatic), isComputed ? property : t().stringLiteral(property.name), value, t().thisExpression()]);
  }

  getSuperProperty(property, isComputed) {
    let thisExpr = t().thisExpression();

    if (this.inConstructor) {
      thisExpr = t().callExpression(this.file.addHelper("assertThisInitialized"), [thisExpr]);
    }

    return t().callExpression(this.file.addHelper("get"), [getPrototypeOfExpression(this.getObjectRef(), this.isStatic), isComputed ? property : t().stringLiteral(property.name), thisExpr]);
  }

  replace() {
    this.methodPath.traverse(visitor, this);
  }

  getLooseSuperProperty(id, parent) {
    const methodNode = this.methodNode;
    const superRef = this.superRef || t().identifier("Function");

    if (parent.property === id) {
      return;
    } else if (t().isCallExpression(parent, {
      callee: id
    })) {
      return;
    } else if (t().isMemberExpression(parent) && !methodNode.static) {
      return t().memberExpression(t().cloneNode(superRef), t().identifier("prototype"));
    } else {
      return t().cloneNode(superRef);
    }
  }

  looseHandle(path) {
    const node = path.node;

    if (path.isSuper()) {
      return this.getLooseSuperProperty(node, path.parent);
    } else if (path.isCallExpression()) {
      const callee = node.callee;
      if (!t().isMemberExpression(callee)) return;
      if (!t().isSuper(callee.object)) return;
      t().appendToMemberExpression(callee, t().identifier("call"));
      node.arguments.unshift(t().thisExpression());
      return true;
    }
  }

  specHandleAssignmentExpression(ref, path, node) {
    if (node.operator === "=") {
      return this.setSuperProperty(node.left.property, node.right, node.left.computed);
    } else {
      ref = ref || path.scope.generateUidIdentifier("ref");
      return [t().variableDeclaration("var", [t().variableDeclarator(t().cloneNode(ref), t().cloneNode(node.left))]), t().expressionStatement(t().assignmentExpression("=", node.left, t().binaryExpression(node.operator.slice(0, -1), t().cloneNode(ref), node.right)))];
    }
  }

  specHandle(path) {
    let property;
    let computed;
    let args;
    const parent = path.parent;
    const node = path.node;

    if (isIllegalBareSuper(node, parent)) {
      throw path.buildCodeFrameError("Illegal use of bare super");
    }

    if (t().isCallExpression(node)) {
      const callee = node.callee;

      if (t().isSuper(callee)) {
        return;
      } else if (isMemberExpressionSuper(callee)) {
        property = callee.property;
        computed = callee.computed;
        args = node.arguments;
      }
    } else if (t().isMemberExpression(node) && t().isSuper(node.object)) {
      property = node.property;
      computed = node.computed;
    } else if (t().isUpdateExpression(node) && isMemberExpressionSuper(node.argument)) {
      const binary = t().assignmentExpression(node.operator[0] + "=", node.argument, t().numericLiteral(1));

      if (node.prefix) {
        return this.specHandleAssignmentExpression(null, path, binary);
      } else {
        const ref = path.scope.generateUidIdentifier("ref");
        return this.specHandleAssignmentExpression(ref, path, binary).concat(t().expressionStatement(ref));
      }
    } else if (t().isAssignmentExpression(node) && isMemberExpressionSuper(node.left)) {
      return this.specHandleAssignmentExpression(null, path, node);
    }

    if (!property) return;
    const superProperty = this.getSuperProperty(property, computed);

    if (args) {
      return this.optimiseCall(superProperty, args);
    } else {
      return superProperty;
    }
  }

  optimiseCall(callee, args) {
    const thisNode = t().thisExpression();
    HARDCORE_THIS_REF.add(thisNode);
    return (0, _helperOptimiseCallExpression().default)(callee, thisNode, args);
  }

}

exports.default = ReplaceSupers;